//! layout tests for scenarios with large size overlapped files
//!
//! See [crate::layout] module for detailed documentation

use std::sync::Arc;

use data_types::CompactionLevel;
use iox_time::{MockProvider, Time, TimeProvider};

use crate::layouts::{layout_setup_builder, parquet_builder, run_layout_scenario, ONE_MB};

#[tokio::test]
async fn one_l1_overlaps_with_many_l2s() {
    // Simulate a production scenario in which there are two L1 files but one overlaps with three L2 files
    // and their total size > limit 300MB
    // |----------L2.1----------||----------L2.2----------||-----L2.3----|
    //  |----------------------------------------L1.4---------------------------||--L1.5--|

    test_helpers::maybe_start_logging();

    let setup = layout_setup_builder().await.build().await;
    let time_provider = Arc::new(MockProvider::new(Time::from_timestamp(0, 0).unwrap()));

    // L2: 100MB, 100MB, 70MB
    // L1: 100MB, 30MB

    // non-overlapping L2 files
    let mut num_l2_files = 0;
    for (i, sz) in [100, 100, 70].iter().enumerate() {
        let i = i as i64;
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(51 + i * 50)
                    .with_max_time(100 + i * 50)
                    .with_compaction_level(CompactionLevel::Final)
                    .with_file_size_bytes(sz * ONE_MB)
                    .with_max_l0_created_at(time_provider.minutes_into_future(i as u64)),
            )
            .await;
        num_l2_files += 1;
    }

    let l2_max_time: i64 = 100 + (num_l2_files - 1) * 50;
    // Non-overlapping L1 files but the first L1 file overlaps with all L2 files
    for (i, sz) in [100, 30].iter().enumerate() {
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(1 + i as i64 * (l2_max_time + 50))
                    .with_max_time((i as i64 + 1) * (l2_max_time + 50))
                    .with_compaction_level(CompactionLevel::FileNonOverlapped)
                    .with_file_size_bytes(sz * ONE_MB)
                    .with_max_l0_created_at(
                        time_provider.minutes_into_future(num_l2_files as u64 + i as u64 + 1),
                    ),
            )
            .await;
    }

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L1                                                                                                                 "
    - "L1.4[1,250] 240s 100mb   |-------------------L1.4-------------------|                                              "
    - "L1.5[251,500] 300s 30mb                                               |-------------------L1.5-------------------| "
    - "L2                                                                                                                 "
    - "L2.1[51,100] 0ns 100mb            |-L2.1-|                                                                         "
    - "L2.2[101,150] 60s 100mb                    |-L2.2-|                                                                "
    - "L2.3[151,200] 120s 70mb                             |-L2.3-|                                                       "
    - "**** Simulation run 0, type=split(ReduceOverlap)(split_times=[100, 150, 200]). 1 Input Files, 100mb total:"
    - "L1, all files 100mb                                                                                                "
    - "L1.4[1,250] 240s         |------------------------------------------L1.4------------------------------------------|"
    - "**** 4 Output Files (parquet_file_id not yet assigned), 100mb total:"
    - "L1                                                                                                                 "
    - "L1.?[1,100] 240s 40mb    |--------------L1.?---------------|                                                       "
    - "L1.?[101,150] 240s 20mb                                      |-----L1.?------|                                     "
    - "L1.?[151,200] 240s 20mb                                                        |-----L1.?------|                   "
    - "L1.?[201,250] 240s 20mb                                                                          |-----L1.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 1 files: L1.4"
    - "  Creating 4 files"
    - "**** Simulation run 1, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[59, 117]). 4 Input Files, 260mb total:"
    - "L1                                                                                                                 "
    - "L1.6[1,100] 240s 40mb    |--------------------------L1.6---------------------------|                               "
    - "L1.7[101,150] 240s 20mb                                                              |-----------L1.7------------| "
    - "L2                                                                                                                 "
    - "L2.1[51,100] 0ns 100mb                                 |-----------L2.1------------|                               "
    - "L2.2[101,150] 60s 100mb                                                              |-----------L2.2------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 260mb total:"
    - "L2                                                                                                                 "
    - "L2.?[1,59] 240s 102mb    |--------------L2.?---------------|                                                       "
    - "L2.?[60,117] 240s 101mb                                     |--------------L2.?--------------|                     "
    - "L2.?[118,150] 240s 57mb                                                                        |------L2.?-------| "
    - "Committing partition 1:"
    - "  Soft Deleting 4 files: L2.1, L2.2, L1.6, L1.7"
    - "  Creating 3 files"
    - "**** Final Output Files (360mb written)"
    - "L1                                                                                                                 "
    - "L1.5[251,500] 300s 30mb                                               |-------------------L1.5-------------------| "
    - "L1.8[151,200] 240s 20mb                             |-L1.8-|                                                       "
    - "L1.9[201,250] 240s 20mb                                      |-L1.9-|                                              "
    - "L2                                                                                                                 "
    - "L2.3[151,200] 120s 70mb                             |-L2.3-|                                                       "
    - "L2.10[1,59] 240s 102mb   |-L2.10--|                                                                                "
    - "L2.11[60,117] 240s 101mb           |-L2.11--|                                                                      "
    - "L2.12[118,150] 240s 57mb                      |L2.12|                                                              "
    "###
    );
}

#[tokio::test]
async fn many_l1_overlaps_with_many_l2s() {
    // Simulate a production scenario in which many L1 files overlap with three L2 files
    // and their total size > limit 256MB
    // |------------L2.1------------||------------L2.2------------||-----L2.3----|
    // |---L1.4---||---L1.5---||-L1.6-||-L1.7-||--L1.8--||--L1.9--||--L1.10--|                      |--L1.11--|

    test_helpers::maybe_start_logging();

    let setup = layout_setup_builder().await.build().await;
    let time_provider = Arc::new(MockProvider::new(Time::from_timestamp(0, 0).unwrap()));

    // L2: 100MB, 100MB, 70MB
    // L1: 8 files each 13MB to have their total size > 100MB to force the compactor to compact them with overlapped L2 files

    // non-overlapping L2 files
    let mut num_l2_files = 0;
    for (i, sz) in [100, 100, 70].iter().enumerate() {
        let i = i as i64;
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(51 + i * 50)
                    .with_max_time(100 + i * 50)
                    .with_compaction_level(CompactionLevel::Final)
                    .with_file_size_bytes(sz * ONE_MB)
                    .with_max_l0_created_at(time_provider.minutes_into_future(i as u64)),
            )
            .await;
        num_l2_files += 1;
    }

    // Non-overlapping L1 files but they overlap with L2 files as in the diagram above
    let mut num_l1_files = 0;
    for (i, sz) in [13, 13, 13, 13, 13, 13, 13].iter().enumerate() {
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(61 + i as i64 * 15)
                    .with_max_time(75 + i as i64 * 15)
                    .with_compaction_level(CompactionLevel::FileNonOverlapped)
                    .with_file_size_bytes(sz * ONE_MB)
                    .with_max_l0_created_at(
                        time_provider.minutes_into_future(num_l2_files as u64 + i as u64 + 1),
                    ),
            )
            .await;
        num_l1_files += 1;
    }
    // Have the last L1 not overlap with anything
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time((num_l2_files as i64 + 1) * 50 + 1)
                .with_max_time((num_l2_files as i64 + 1) * 50 + 15)
                .with_compaction_level(CompactionLevel::FileNonOverlapped)
                .with_file_size_bytes(13 * ONE_MB)
                .with_max_l0_created_at(
                    time_provider
                        .minutes_into_future(num_l2_files as u64 + num_l1_files as u64 + 1),
                ),
        )
        .await;

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L1                                                                                                                 "
    - "L1.4[61,75] 240s 13mb         |L1.4-|                                                                              "
    - "L1.5[76,90] 300s 13mb                 |L1.5-|                                                                      "
    - "L1.6[91,105] 360s 13mb                        |L1.6-|                                                              "
    - "L1.7[106,120] 420s 13mb                                |L1.7-|                                                     "
    - "L1.8[121,135] 480s 13mb                                        |L1.8-|                                             "
    - "L1.9[136,150] 540s 13mb                                                |L1.9-|                                     "
    - "L1.10[151,165] 600s 13mb                                                       |L1.10|                             "
    - "L1.11[201,215] 660s 13mb                                                                                   |L1.11| "
    - "L2                                                                                                                 "
    - "L2.1[51,100] 0ns 100mb   |----------L2.1----------|                                                                "
    - "L2.2[101,150] 60s 100mb                             |----------L2.2----------|                                     "
    - "L2.3[151,200] 120s 70mb                                                        |----------L2.3----------|          "
    - "**** Simulation run 0, type=split(ReduceOverlap)(split_times=[100]). 1 Input Files, 13mb total:"
    - "L1, all files 13mb                                                                                                 "
    - "L1.6[91,105] 360s        |------------------------------------------L1.6------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 13mb total:"
    - "L1                                                                                                                 "
    - "L1.?[91,100] 360s 9mb    |-------------------------L1.?--------------------------|                                 "
    - "L1.?[101,105] 360s 4mb                                                                   |---------L1.?----------| "
    - "Committing partition 1:"
    - "  Soft Deleting 1 files: L1.6"
    - "  Creating 2 files"
    - "**** Simulation run 1, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[87, 123]). 9 Input Files, 278mb total:"
    - "L1                                                                                                                 "
    - "L1.4[61,75] 240s 13mb             |---L1.4---|                                                                     "
    - "L1.5[76,90] 300s 13mb                          |---L1.5---|                                                        "
    - "L1.12[91,100] 360s 9mb                                       |L1.12-|                                              "
    - "L1.13[101,105] 360s 4mb                                               |L1.13|                                      "
    - "L1.7[106,120] 420s 13mb                                                   |---L1.7---|                             "
    - "L1.8[121,135] 480s 13mb                                                                 |---L1.8---|               "
    - "L1.9[136,150] 540s 13mb                                                                               |---L1.9---| "
    - "L2                                                                                                                 "
    - "L2.1[51,100] 0ns 100mb   |-------------------L2.1-------------------|                                              "
    - "L2.2[101,150] 60s 100mb                                               |-------------------L2.2-------------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 278mb total:"
    - "L2                                                                                                                 "
    - "L2.?[51,87] 540s 103mb   |-------------L2.?-------------|                                                          "
    - "L2.?[88,123] 540s 100mb                                   |------------L2.?-------------|                          "
    - "L2.?[124,150] 540s 75mb                                                                    |--------L2.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 9 files: L2.1, L2.2, L1.4, L1.5, L1.7, L1.8, L1.9, L1.12, L1.13"
    - "  Creating 3 files"
    - "**** Final Output Files (291mb written)"
    - "L1                                                                                                                 "
    - "L1.10[151,165] 600s 13mb                                                       |L1.10|                             "
    - "L1.11[201,215] 660s 13mb                                                                                   |L1.11| "
    - "L2                                                                                                                 "
    - "L2.3[151,200] 120s 70mb                                                        |----------L2.3----------|          "
    - "L2.14[51,87] 540s 103mb  |------L2.14------|                                                                       "
    - "L2.15[88,123] 540s 100mb                     |------L2.15------|                                                   "
    - "L2.16[124,150] 540s 75mb                                         |---L2.16----|                                    "
    "###
    );
}

#[tokio::test]
async fn many_good_size_l0_files() {
    test_helpers::maybe_start_logging();

    // Scenario when we have a lot of L0 files because the compactor cannot keep up with ingesters
    // and each files is a good size, 2MB, which lead to total size of 200 files (limit num files)
    // greater than limit size (256MB)

    let two_mb = 2 * ONE_MB;

    let setup = layout_setup_builder().await.build().await;

    let num_files = (24 * 60) / 5; // 288 files

    for i in 0..num_files {
        let i = i as i64;
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(i)
                    .with_max_time(i + 1)
                    .with_compaction_level(CompactionLevel::Initial)
                    .with_max_l0_created_at(Time::from_timestamp_nanos(i + 1))
                    .with_file_size_bytes(two_mb),
            )
            .await;
    }

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 2mb                                                                                                  "
    - "L0.1[0,1] 1ns            |L0.1|                                                                                    "
    - "L0.2[1,2] 2ns            |L0.2|                                                                                    "
    - "L0.3[2,3] 3ns            |L0.3|                                                                                    "
    - "L0.4[3,4] 4ns            |L0.4|                                                                                    "
    - "L0.5[4,5] 5ns             |L0.5|                                                                                   "
    - "L0.6[5,6] 6ns             |L0.6|                                                                                   "
    - "L0.7[6,7] 7ns             |L0.7|                                                                                   "
    - "L0.8[7,8] 8ns              |L0.8|                                                                                  "
    - "L0.9[8,9] 9ns              |L0.9|                                                                                  "
    - "L0.10[9,10] 10ns           |L0.10|                                                                                 "
    - "L0.11[10,11] 11ns           |L0.11|                                                                                "
    - "L0.12[11,12] 12ns           |L0.12|                                                                                "
    - "L0.13[12,13] 13ns           |L0.13|                                                                                "
    - "L0.14[13,14] 14ns            |L0.14|                                                                               "
    - "L0.15[14,15] 15ns            |L0.15|                                                                               "
    - "L0.16[15,16] 16ns            |L0.16|                                                                               "
    - "L0.17[16,17] 17ns             |L0.17|                                                                              "
    - "L0.18[17,18] 18ns             |L0.18|                                                                              "
    - "L0.19[18,19] 19ns             |L0.19|                                                                              "
    - "L0.20[19,20] 20ns             |L0.20|                                                                              "
    - "L0.21[20,21] 21ns              |L0.21|                                                                             "
    - "L0.22[21,22] 22ns              |L0.22|                                                                             "
    - "L0.23[22,23] 23ns              |L0.23|                                                                             "
    - "L0.24[23,24] 24ns               |L0.24|                                                                            "
    - "L0.25[24,25] 25ns               |L0.25|                                                                            "
    - "L0.26[25,26] 26ns               |L0.26|                                                                            "
    - "L0.27[26,27] 27ns                |L0.27|                                                                           "
    - "L0.28[27,28] 28ns                |L0.28|                                                                           "
    - "L0.29[28,29] 29ns                |L0.29|                                                                           "
    - "L0.30[29,30] 30ns                 |L0.30|                                                                          "
    - "L0.31[30,31] 31ns                 |L0.31|                                                                          "
    - "L0.32[31,32] 32ns                 |L0.32|                                                                          "
    - "L0.33[32,33] 33ns                  |L0.33|                                                                         "
    - "L0.34[33,34] 34ns                  |L0.34|                                                                         "
    - "L0.35[34,35] 35ns                  |L0.35|                                                                         "
    - "L0.36[35,36] 36ns                  |L0.36|                                                                         "
    - "L0.37[36,37] 37ns                   |L0.37|                                                                        "
    - "L0.38[37,38] 38ns                   |L0.38|                                                                        "
    - "L0.39[38,39] 39ns                   |L0.39|                                                                        "
    - "L0.40[39,40] 40ns                    |L0.40|                                                                       "
    - "L0.41[40,41] 41ns                    |L0.41|                                                                       "
    - "L0.42[41,42] 42ns                    |L0.42|                                                                       "
    - "L0.43[42,43] 43ns                     |L0.43|                                                                      "
    - "L0.44[43,44] 44ns                     |L0.44|                                                                      "
    - "L0.45[44,45] 45ns                     |L0.45|                                                                      "
    - "L0.46[45,46] 46ns                      |L0.46|                                                                     "
    - "L0.47[46,47] 47ns                      |L0.47|                                                                     "
    - "L0.48[47,48] 48ns                      |L0.48|                                                                     "
    - "L0.49[48,49] 49ns                       |L0.49|                                                                    "
    - "L0.50[49,50] 50ns                       |L0.50|                                                                    "
    - "L0.51[50,51] 51ns                       |L0.51|                                                                    "
    - "L0.52[51,52] 52ns                       |L0.52|                                                                    "
    - "L0.53[52,53] 53ns                        |L0.53|                                                                   "
    - "L0.54[53,54] 54ns                        |L0.54|                                                                   "
    - "L0.55[54,55] 55ns                        |L0.55|                                                                   "
    - "L0.56[55,56] 56ns                         |L0.56|                                                                  "
    - "L0.57[56,57] 57ns                         |L0.57|                                                                  "
    - "L0.58[57,58] 58ns                         |L0.58|                                                                  "
    - "L0.59[58,59] 59ns                          |L0.59|                                                                 "
    - "L0.60[59,60] 60ns                          |L0.60|                                                                 "
    - "L0.61[60,61] 61ns                          |L0.61|                                                                 "
    - "L0.62[61,62] 62ns                           |L0.62|                                                                "
    - "L0.63[62,63] 63ns                           |L0.63|                                                                "
    - "L0.64[63,64] 64ns                           |L0.64|                                                                "
    - "L0.65[64,65] 65ns                            |L0.65|                                                               "
    - "L0.66[65,66] 66ns                            |L0.66|                                                               "
    - "L0.67[66,67] 67ns                            |L0.67|                                                               "
    - "L0.68[67,68] 68ns                            |L0.68|                                                               "
    - "L0.69[68,69] 69ns                             |L0.69|                                                              "
    - "L0.70[69,70] 70ns                             |L0.70|                                                              "
    - "L0.71[70,71] 71ns                             |L0.71|                                                              "
    - "L0.72[71,72] 72ns                              |L0.72|                                                             "
    - "L0.73[72,73] 73ns                              |L0.73|                                                             "
    - "L0.74[73,74] 74ns                              |L0.74|                                                             "
    - "L0.75[74,75] 75ns                               |L0.75|                                                            "
    - "L0.76[75,76] 76ns                               |L0.76|                                                            "
    - "L0.77[76,77] 77ns                               |L0.77|                                                            "
    - "L0.78[77,78] 78ns                                |L0.78|                                                           "
    - "L0.79[78,79] 79ns                                |L0.79|                                                           "
    - "L0.80[79,80] 80ns                                |L0.80|                                                           "
    - "L0.81[80,81] 81ns                                 |L0.81|                                                          "
    - "L0.82[81,82] 82ns                                 |L0.82|                                                          "
    - "L0.83[82,83] 83ns                                 |L0.83|                                                          "
    - "L0.84[83,84] 84ns                                 |L0.84|                                                          "
    - "L0.85[84,85] 85ns                                  |L0.85|                                                         "
    - "L0.86[85,86] 86ns                                  |L0.86|                                                         "
    - "L0.87[86,87] 87ns                                  |L0.87|                                                         "
    - "L0.88[87,88] 88ns                                   |L0.88|                                                        "
    - "L0.89[88,89] 89ns                                   |L0.89|                                                        "
    - "L0.90[89,90] 90ns                                   |L0.90|                                                        "
    - "L0.91[90,91] 91ns                                    |L0.91|                                                       "
    - "L0.92[91,92] 92ns                                    |L0.92|                                                       "
    - "L0.93[92,93] 93ns                                    |L0.93|                                                       "
    - "L0.94[93,94] 94ns                                     |L0.94|                                                      "
    - "L0.95[94,95] 95ns                                     |L0.95|                                                      "
    - "L0.96[95,96] 96ns                                     |L0.96|                                                      "
    - "L0.97[96,97] 97ns                                      |L0.97|                                                     "
    - "L0.98[97,98] 98ns                                      |L0.98|                                                     "
    - "L0.99[98,99] 99ns                                      |L0.99|                                                     "
    - "L0.100[99,100] 100ns                                   |L0.100|                                                    "
    - "L0.101[100,101] 101ns                                   |L0.101|                                                   "
    - "L0.102[101,102] 102ns                                   |L0.102|                                                   "
    - "L0.103[102,103] 103ns                                   |L0.103|                                                   "
    - "L0.104[103,104] 104ns                                    |L0.104|                                                  "
    - "L0.105[104,105] 105ns                                    |L0.105|                                                  "
    - "L0.106[105,106] 106ns                                    |L0.106|                                                  "
    - "L0.107[106,107] 107ns                                     |L0.107|                                                 "
    - "L0.108[107,108] 108ns                                     |L0.108|                                                 "
    - "L0.109[108,109] 109ns                                     |L0.109|                                                 "
    - "L0.110[109,110] 110ns                                      |L0.110|                                                "
    - "L0.111[110,111] 111ns                                      |L0.111|                                                "
    - "L0.112[111,112] 112ns                                      |L0.112|                                                "
    - "L0.113[112,113] 113ns                                       |L0.113|                                               "
    - "L0.114[113,114] 114ns                                       |L0.114|                                               "
    - "L0.115[114,115] 115ns                                       |L0.115|                                               "
    - "L0.116[115,116] 116ns                                       |L0.116|                                               "
    - "L0.117[116,117] 117ns                                        |L0.117|                                              "
    - "L0.118[117,118] 118ns                                        |L0.118|                                              "
    - "L0.119[118,119] 119ns                                        |L0.119|                                              "
    - "L0.120[119,120] 120ns                                         |L0.120|                                             "
    - "L0.121[120,121] 121ns                                         |L0.121|                                             "
    - "L0.122[121,122] 122ns                                         |L0.122|                                             "
    - "L0.123[122,123] 123ns                                          |L0.123|                                            "
    - "L0.124[123,124] 124ns                                          |L0.124|                                            "
    - "L0.125[124,125] 125ns                                          |L0.125|                                            "
    - "L0.126[125,126] 126ns                                           |L0.126|                                           "
    - "L0.127[126,127] 127ns                                           |L0.127|                                           "
    - "L0.128[127,128] 128ns                                           |L0.128|                                           "
    - "L0.129[128,129] 129ns                                            |L0.129|                                          "
    - "L0.130[129,130] 130ns                                            |L0.130|                                          "
    - "L0.131[130,131] 131ns                                            |L0.131|                                          "
    - "L0.132[131,132] 132ns                                            |L0.132|                                          "
    - "L0.133[132,133] 133ns                                             |L0.133|                                         "
    - "L0.134[133,134] 134ns                                             |L0.134|                                         "
    - "L0.135[134,135] 135ns                                             |L0.135|                                         "
    - "L0.136[135,136] 136ns                                              |L0.136|                                        "
    - "L0.137[136,137] 137ns                                              |L0.137|                                        "
    - "L0.138[137,138] 138ns                                              |L0.138|                                        "
    - "L0.139[138,139] 139ns                                               |L0.139|                                       "
    - "L0.140[139,140] 140ns                                               |L0.140|                                       "
    - "L0.141[140,141] 141ns                                               |L0.141|                                       "
    - "L0.142[141,142] 142ns                                                |L0.142|                                      "
    - "L0.143[142,143] 143ns                                                |L0.143|                                      "
    - "L0.144[143,144] 144ns                                                |L0.144|                                      "
    - "L0.145[144,145] 145ns                                                 |L0.145|                                     "
    - "L0.146[145,146] 146ns                                                 |L0.146|                                     "
    - "L0.147[146,147] 147ns                                                 |L0.147|                                     "
    - "L0.148[147,148] 148ns                                                 |L0.148|                                     "
    - "L0.149[148,149] 149ns                                                  |L0.149|                                    "
    - "L0.150[149,150] 150ns                                                  |L0.150|                                    "
    - "L0.151[150,151] 151ns                                                  |L0.151|                                    "
    - "L0.152[151,152] 152ns                                                   |L0.152|                                   "
    - "L0.153[152,153] 153ns                                                   |L0.153|                                   "
    - "L0.154[153,154] 154ns                                                   |L0.154|                                   "
    - "L0.155[154,155] 155ns                                                    |L0.155|                                  "
    - "L0.156[155,156] 156ns                                                    |L0.156|                                  "
    - "L0.157[156,157] 157ns                                                    |L0.157|                                  "
    - "L0.158[157,158] 158ns                                                     |L0.158|                                 "
    - "L0.159[158,159] 159ns                                                     |L0.159|                                 "
    - "L0.160[159,160] 160ns                                                     |L0.160|                                 "
    - "L0.161[160,161] 161ns                                                      |L0.161|                                "
    - "L0.162[161,162] 162ns                                                      |L0.162|                                "
    - "L0.163[162,163] 163ns                                                      |L0.163|                                "
    - "L0.164[163,164] 164ns                                                      |L0.164|                                "
    - "L0.165[164,165] 165ns                                                       |L0.165|                               "
    - "L0.166[165,166] 166ns                                                       |L0.166|                               "
    - "L0.167[166,167] 167ns                                                       |L0.167|                               "
    - "L0.168[167,168] 168ns                                                        |L0.168|                              "
    - "L0.169[168,169] 169ns                                                        |L0.169|                              "
    - "L0.170[169,170] 170ns                                                        |L0.170|                              "
    - "L0.171[170,171] 171ns                                                         |L0.171|                             "
    - "L0.172[171,172] 172ns                                                         |L0.172|                             "
    - "L0.173[172,173] 173ns                                                         |L0.173|                             "
    - "L0.174[173,174] 174ns                                                          |L0.174|                            "
    - "L0.175[174,175] 175ns                                                          |L0.175|                            "
    - "L0.176[175,176] 176ns                                                          |L0.176|                            "
    - "L0.177[176,177] 177ns                                                           |L0.177|                           "
    - "L0.178[177,178] 178ns                                                           |L0.178|                           "
    - "L0.179[178,179] 179ns                                                           |L0.179|                           "
    - "L0.180[179,180] 180ns                                                           |L0.180|                           "
    - "L0.181[180,181] 181ns                                                            |L0.181|                          "
    - "L0.182[181,182] 182ns                                                            |L0.182|                          "
    - "L0.183[182,183] 183ns                                                            |L0.183|                          "
    - "L0.184[183,184] 184ns                                                             |L0.184|                         "
    - "L0.185[184,185] 185ns                                                             |L0.185|                         "
    - "L0.186[185,186] 186ns                                                             |L0.186|                         "
    - "L0.187[186,187] 187ns                                                              |L0.187|                        "
    - "L0.188[187,188] 188ns                                                              |L0.188|                        "
    - "L0.189[188,189] 189ns                                                              |L0.189|                        "
    - "L0.190[189,190] 190ns                                                               |L0.190|                       "
    - "L0.191[190,191] 191ns                                                               |L0.191|                       "
    - "L0.192[191,192] 192ns                                                               |L0.192|                       "
    - "L0.193[192,193] 193ns                                                                |L0.193|                      "
    - "L0.194[193,194] 194ns                                                                |L0.194|                      "
    - "L0.195[194,195] 195ns                                                                |L0.195|                      "
    - "L0.196[195,196] 196ns                                                                |L0.196|                      "
    - "L0.197[196,197] 197ns                                                                 |L0.197|                     "
    - "L0.198[197,198] 198ns                                                                 |L0.198|                     "
    - "L0.199[198,199] 199ns                                                                 |L0.199|                     "
    - "L0.200[199,200] 200ns                                                                  |L0.200|                    "
    - "L0.201[200,201] 201ns                                                                  |L0.201|                    "
    - "L0.202[201,202] 202ns                                                                  |L0.202|                    "
    - "L0.203[202,203] 203ns                                                                   |L0.203|                   "
    - "L0.204[203,204] 204ns                                                                   |L0.204|                   "
    - "L0.205[204,205] 205ns                                                                   |L0.205|                   "
    - "L0.206[205,206] 206ns                                                                    |L0.206|                  "
    - "L0.207[206,207] 207ns                                                                    |L0.207|                  "
    - "L0.208[207,208] 208ns                                                                    |L0.208|                  "
    - "L0.209[208,209] 209ns                                                                     |L0.209|                 "
    - "L0.210[209,210] 210ns                                                                     |L0.210|                 "
    - "L0.211[210,211] 211ns                                                                     |L0.211|                 "
    - "L0.212[211,212] 212ns                                                                     |L0.212|                 "
    - "L0.213[212,213] 213ns                                                                      |L0.213|                "
    - "L0.214[213,214] 214ns                                                                      |L0.214|                "
    - "L0.215[214,215] 215ns                                                                      |L0.215|                "
    - "L0.216[215,216] 216ns                                                                       |L0.216|               "
    - "L0.217[216,217] 217ns                                                                       |L0.217|               "
    - "L0.218[217,218] 218ns                                                                       |L0.218|               "
    - "L0.219[218,219] 219ns                                                                        |L0.219|              "
    - "L0.220[219,220] 220ns                                                                        |L0.220|              "
    - "L0.221[220,221] 221ns                                                                        |L0.221|              "
    - "L0.222[221,222] 222ns                                                                         |L0.222|             "
    - "L0.223[222,223] 223ns                                                                         |L0.223|             "
    - "L0.224[223,224] 224ns                                                                         |L0.224|             "
    - "L0.225[224,225] 225ns                                                                          |L0.225|            "
    - "L0.226[225,226] 226ns                                                                          |L0.226|            "
    - "L0.227[226,227] 227ns                                                                          |L0.227|            "
    - "L0.228[227,228] 228ns                                                                          |L0.228|            "
    - "L0.229[228,229] 229ns                                                                           |L0.229|           "
    - "L0.230[229,230] 230ns                                                                           |L0.230|           "
    - "L0.231[230,231] 231ns                                                                           |L0.231|           "
    - "L0.232[231,232] 232ns                                                                            |L0.232|          "
    - "L0.233[232,233] 233ns                                                                            |L0.233|          "
    - "L0.234[233,234] 234ns                                                                            |L0.234|          "
    - "L0.235[234,235] 235ns                                                                             |L0.235|         "
    - "L0.236[235,236] 236ns                                                                             |L0.236|         "
    - "L0.237[236,237] 237ns                                                                             |L0.237|         "
    - "L0.238[237,238] 238ns                                                                              |L0.238|        "
    - "L0.239[238,239] 239ns                                                                              |L0.239|        "
    - "L0.240[239,240] 240ns                                                                              |L0.240|        "
    - "L0.241[240,241] 241ns                                                                               |L0.241|       "
    - "L0.242[241,242] 242ns                                                                               |L0.242|       "
    - "L0.243[242,243] 243ns                                                                               |L0.243|       "
    - "L0.244[243,244] 244ns                                                                               |L0.244|       "
    - "L0.245[244,245] 245ns                                                                                |L0.245|      "
    - "L0.246[245,246] 246ns                                                                                |L0.246|      "
    - "L0.247[246,247] 247ns                                                                                |L0.247|      "
    - "L0.248[247,248] 248ns                                                                                 |L0.248|     "
    - "L0.249[248,249] 249ns                                                                                 |L0.249|     "
    - "L0.250[249,250] 250ns                                                                                 |L0.250|     "
    - "L0.251[250,251] 251ns                                                                                  |L0.251|    "
    - "L0.252[251,252] 252ns                                                                                  |L0.252|    "
    - "L0.253[252,253] 253ns                                                                                  |L0.253|    "
    - "L0.254[253,254] 254ns                                                                                   |L0.254|   "
    - "L0.255[254,255] 255ns                                                                                   |L0.255|   "
    - "L0.256[255,256] 256ns                                                                                   |L0.256|   "
    - "L0.257[256,257] 257ns                                                                                    |L0.257|  "
    - "L0.258[257,258] 258ns                                                                                    |L0.258|  "
    - "L0.259[258,259] 259ns                                                                                    |L0.259|  "
    - "L0.260[259,260] 260ns                                                                                    |L0.260|  "
    - "L0.261[260,261] 261ns                                                                                     |L0.261| "
    - "L0.262[261,262] 262ns                                                                                     |L0.262| "
    - "L0.263[262,263] 263ns                                                                                     |L0.263| "
    - "L0.264[263,264] 264ns                                                                                      |L0.264|"
    - "L0.265[264,265] 265ns                                                                                      |L0.265|"
    - "L0.266[265,266] 266ns                                                                                      |L0.266|"
    - "L0.267[266,267] 267ns                                                                                       |L0.267|"
    - "L0.268[267,268] 268ns                                                                                       |L0.268|"
    - "L0.269[268,269] 269ns                                                                                       |L0.269|"
    - "L0.270[269,270] 270ns                                                                                        |L0.270|"
    - "L0.271[270,271] 271ns                                                                                        |L0.271|"
    - "L0.272[271,272] 272ns                                                                                        |L0.272|"
    - "L0.273[272,273] 273ns                                                                                         |L0.273|"
    - "L0.274[273,274] 274ns                                                                                         |L0.274|"
    - "L0.275[274,275] 275ns                                                                                         |L0.275|"
    - "L0.276[275,276] 276ns                                                                                         |L0.276|"
    - "L0.277[276,277] 277ns                                                                                          |L0.277|"
    - "L0.278[277,278] 278ns                                                                                          |L0.278|"
    - "L0.279[278,279] 279ns                                                                                          |L0.279|"
    - "L0.280[279,280] 280ns                                                                                           |L0.280|"
    - "L0.281[280,281] 281ns                                                                                           |L0.281|"
    - "L0.282[281,282] 282ns                                                                                           |L0.282|"
    - "L0.283[282,283] 283ns                                                                                            |L0.283|"
    - "L0.284[283,284] 284ns                                                                                            |L0.284|"
    - "L0.285[284,285] 285ns                                                                                            |L0.285|"
    - "L0.286[285,286] 286ns                                                                                             |L0.286|"
    - "L0.287[286,287] 287ns                                                                                             |L0.287|"
    - "L0.288[287,288] 288ns                                                                                             |L0.288|"
    - "**** Simulation run 0, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[50, 100]). 150 Input Files, 300mb total:"
    - "L0, all files 2mb                                                                                                  "
    - "L0.1[0,1] 1ns            |L0.1|                                                                                    "
    - "L0.2[1,2] 2ns            |L0.2|                                                                                    "
    - "L0.3[2,3] 3ns             |L0.3|                                                                                   "
    - "L0.4[3,4] 4ns             |L0.4|                                                                                   "
    - "L0.5[4,5] 5ns              |L0.5|                                                                                  "
    - "L0.6[5,6] 6ns               |L0.6|                                                                                 "
    - "L0.7[6,7] 7ns               |L0.7|                                                                                 "
    - "L0.8[7,8] 8ns                |L0.8|                                                                                "
    - "L0.9[8,9] 9ns                |L0.9|                                                                                "
    - "L0.10[9,10] 10ns              |L0.10|                                                                              "
    - "L0.11[10,11] 11ns              |L0.11|                                                                             "
    - "L0.12[11,12] 12ns              |L0.12|                                                                             "
    - "L0.13[12,13] 13ns               |L0.13|                                                                            "
    - "L0.14[13,14] 14ns               |L0.14|                                                                            "
    - "L0.15[14,15] 15ns                |L0.15|                                                                           "
    - "L0.16[15,16] 16ns                 |L0.16|                                                                          "
    - "L0.17[16,17] 17ns                 |L0.17|                                                                          "
    - "L0.18[17,18] 18ns                  |L0.18|                                                                         "
    - "L0.19[18,19] 19ns                  |L0.19|                                                                         "
    - "L0.20[19,20] 20ns                   |L0.20|                                                                        "
    - "L0.21[20,21] 21ns                    |L0.21|                                                                       "
    - "L0.22[21,22] 22ns                    |L0.22|                                                                       "
    - "L0.23[22,23] 23ns                     |L0.23|                                                                      "
    - "L0.24[23,24] 24ns                     |L0.24|                                                                      "
    - "L0.25[24,25] 25ns                      |L0.25|                                                                     "
    - "L0.26[25,26] 26ns                       |L0.26|                                                                    "
    - "L0.27[26,27] 27ns                       |L0.27|                                                                    "
    - "L0.28[27,28] 28ns                        |L0.28|                                                                   "
    - "L0.29[28,29] 29ns                        |L0.29|                                                                   "
    - "L0.30[29,30] 30ns                         |L0.30|                                                                  "
    - "L0.31[30,31] 31ns                          |L0.31|                                                                 "
    - "L0.32[31,32] 32ns                          |L0.32|                                                                 "
    - "L0.33[32,33] 33ns                           |L0.33|                                                                "
    - "L0.34[33,34] 34ns                           |L0.34|                                                                "
    - "L0.35[34,35] 35ns                            |L0.35|                                                               "
    - "L0.36[35,36] 36ns                             |L0.36|                                                              "
    - "L0.37[36,37] 37ns                             |L0.37|                                                              "
    - "L0.38[37,38] 38ns                              |L0.38|                                                             "
    - "L0.39[38,39] 39ns                              |L0.39|                                                             "
    - "L0.40[39,40] 40ns                               |L0.40|                                                            "
    - "L0.41[40,41] 41ns                                |L0.41|                                                           "
    - "L0.42[41,42] 42ns                                |L0.42|                                                           "
    - "L0.43[42,43] 43ns                                 |L0.43|                                                          "
    - "L0.44[43,44] 44ns                                 |L0.44|                                                          "
    - "L0.45[44,45] 45ns                                  |L0.45|                                                         "
    - "L0.46[45,46] 46ns                                   |L0.46|                                                        "
    - "L0.47[46,47] 47ns                                   |L0.47|                                                        "
    - "L0.48[47,48] 48ns                                    |L0.48|                                                       "
    - "L0.49[48,49] 49ns                                    |L0.49|                                                       "
    - "L0.50[49,50] 50ns                                     |L0.50|                                                      "
    - "L0.51[50,51] 51ns                                      |L0.51|                                                     "
    - "L0.52[51,52] 52ns                                      |L0.52|                                                     "
    - "L0.53[52,53] 53ns                                       |L0.53|                                                    "
    - "L0.54[53,54] 54ns                                       |L0.54|                                                    "
    - "L0.55[54,55] 55ns                                        |L0.55|                                                   "
    - "L0.56[55,56] 56ns                                         |L0.56|                                                  "
    - "L0.57[56,57] 57ns                                         |L0.57|                                                  "
    - "L0.58[57,58] 58ns                                          |L0.58|                                                 "
    - "L0.59[58,59] 59ns                                          |L0.59|                                                 "
    - "L0.60[59,60] 60ns                                           |L0.60|                                                "
    - "L0.61[60,61] 61ns                                            |L0.61|                                               "
    - "L0.62[61,62] 62ns                                            |L0.62|                                               "
    - "L0.63[62,63] 63ns                                             |L0.63|                                              "
    - "L0.64[63,64] 64ns                                             |L0.64|                                              "
    - "L0.65[64,65] 65ns                                              |L0.65|                                             "
    - "L0.66[65,66] 66ns                                               |L0.66|                                            "
    - "L0.67[66,67] 67ns                                               |L0.67|                                            "
    - "L0.68[67,68] 68ns                                                |L0.68|                                           "
    - "L0.69[68,69] 69ns                                                |L0.69|                                           "
    - "L0.70[69,70] 70ns                                                 |L0.70|                                          "
    - "L0.71[70,71] 71ns                                                  |L0.71|                                         "
    - "L0.72[71,72] 72ns                                                  |L0.72|                                         "
    - "L0.73[72,73] 73ns                                                   |L0.73|                                        "
    - "L0.74[73,74] 74ns                                                   |L0.74|                                        "
    - "L0.75[74,75] 75ns                                                    |L0.75|                                       "
    - "L0.76[75,76] 76ns                                                     |L0.76|                                      "
    - "L0.77[76,77] 77ns                                                     |L0.77|                                      "
    - "L0.78[77,78] 78ns                                                      |L0.78|                                     "
    - "L0.79[78,79] 79ns                                                      |L0.79|                                     "
    - "L0.80[79,80] 80ns                                                       |L0.80|                                    "
    - "L0.81[80,81] 81ns                                                        |L0.81|                                   "
    - "L0.82[81,82] 82ns                                                        |L0.82|                                   "
    - "L0.83[82,83] 83ns                                                         |L0.83|                                  "
    - "L0.84[83,84] 84ns                                                         |L0.84|                                  "
    - "L0.85[84,85] 85ns                                                          |L0.85|                                 "
    - "L0.86[85,86] 86ns                                                           |L0.86|                                "
    - "L0.87[86,87] 87ns                                                           |L0.87|                                "
    - "L0.88[87,88] 88ns                                                            |L0.88|                               "
    - "L0.89[88,89] 89ns                                                            |L0.89|                               "
    - "L0.90[89,90] 90ns                                                             |L0.90|                              "
    - "L0.91[90,91] 91ns                                                              |L0.91|                             "
    - "L0.92[91,92] 92ns                                                              |L0.92|                             "
    - "L0.93[92,93] 93ns                                                               |L0.93|                            "
    - "L0.94[93,94] 94ns                                                               |L0.94|                            "
    - "L0.95[94,95] 95ns                                                                |L0.95|                           "
    - "L0.96[95,96] 96ns                                                                 |L0.96|                          "
    - "L0.97[96,97] 97ns                                                                 |L0.97|                          "
    - "L0.98[97,98] 98ns                                                                  |L0.98|                         "
    - "L0.99[98,99] 99ns                                                                  |L0.99|                         "
    - "L0.100[99,100] 100ns                                                                |L0.100|                       "
    - "L0.101[100,101] 101ns                                                                |L0.101|                      "
    - "L0.102[101,102] 102ns                                                                |L0.102|                      "
    - "L0.103[102,103] 103ns                                                                 |L0.103|                     "
    - "L0.104[103,104] 104ns                                                                 |L0.104|                     "
    - "L0.105[104,105] 105ns                                                                  |L0.105|                    "
    - "L0.106[105,106] 106ns                                                                   |L0.106|                   "
    - "L0.107[106,107] 107ns                                                                   |L0.107|                   "
    - "L0.108[107,108] 108ns                                                                    |L0.108|                  "
    - "L0.109[108,109] 109ns                                                                    |L0.109|                  "
    - "L0.110[109,110] 110ns                                                                     |L0.110|                 "
    - "L0.111[110,111] 111ns                                                                      |L0.111|                "
    - "L0.112[111,112] 112ns                                                                      |L0.112|                "
    - "L0.113[112,113] 113ns                                                                       |L0.113|               "
    - "L0.114[113,114] 114ns                                                                       |L0.114|               "
    - "L0.115[114,115] 115ns                                                                        |L0.115|              "
    - "L0.116[115,116] 116ns                                                                         |L0.116|             "
    - "L0.117[116,117] 117ns                                                                         |L0.117|             "
    - "L0.118[117,118] 118ns                                                                          |L0.118|            "
    - "L0.119[118,119] 119ns                                                                          |L0.119|            "
    - "L0.120[119,120] 120ns                                                                           |L0.120|           "
    - "L0.121[120,121] 121ns                                                                            |L0.121|          "
    - "L0.122[121,122] 122ns                                                                            |L0.122|          "
    - "L0.123[122,123] 123ns                                                                             |L0.123|         "
    - "L0.124[123,124] 124ns                                                                             |L0.124|         "
    - "L0.125[124,125] 125ns                                                                              |L0.125|        "
    - "L0.126[125,126] 126ns                                                                               |L0.126|       "
    - "L0.127[126,127] 127ns                                                                               |L0.127|       "
    - "L0.128[127,128] 128ns                                                                                |L0.128|      "
    - "L0.129[128,129] 129ns                                                                                |L0.129|      "
    - "L0.130[129,130] 130ns                                                                                 |L0.130|     "
    - "L0.131[130,131] 131ns                                                                                  |L0.131|    "
    - "L0.132[131,132] 132ns                                                                                  |L0.132|    "
    - "L0.133[132,133] 133ns                                                                                   |L0.133|   "
    - "L0.134[133,134] 134ns                                                                                   |L0.134|   "
    - "L0.135[134,135] 135ns                                                                                    |L0.135|  "
    - "L0.136[135,136] 136ns                                                                                     |L0.136| "
    - "L0.137[136,137] 137ns                                                                                     |L0.137| "
    - "L0.138[137,138] 138ns                                                                                      |L0.138|"
    - "L0.139[138,139] 139ns                                                                                      |L0.139|"
    - "L0.140[139,140] 140ns                                                                                       |L0.140|"
    - "L0.141[140,141] 141ns                                                                                        |L0.141|"
    - "L0.142[141,142] 142ns                                                                                        |L0.142|"
    - "L0.143[142,143] 143ns                                                                                         |L0.143|"
    - "L0.144[143,144] 144ns                                                                                         |L0.144|"
    - "L0.145[144,145] 145ns                                                                                          |L0.145|"
    - "L0.146[145,146] 146ns                                                                                           |L0.146|"
    - "L0.147[146,147] 147ns                                                                                           |L0.147|"
    - "L0.148[147,148] 148ns                                                                                            |L0.148|"
    - "L0.149[148,149] 149ns                                                                                            |L0.149|"
    - "L0.150[149,150] 150ns                                                                                             |L0.150|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 300mb total:"
    - "L1                                                                                                                 "
    - "L1.?[0,50] 150ns 101mb   |------------L1.?------------|                                                            "
    - "L1.?[51,100] 150ns 99mb                                |-----------L1.?------------|                               "
    - "L1.?[101,150] 150ns 99mb                                                             |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 150 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10, L0.11, L0.12, L0.13, L0.14, L0.15, L0.16, L0.17, L0.18, L0.19, L0.20, L0.21, L0.22, L0.23, L0.24, L0.25, L0.26, L0.27, L0.28, L0.29, L0.30, L0.31, L0.32, L0.33, L0.34, L0.35, L0.36, L0.37, L0.38, L0.39, L0.40, L0.41, L0.42, L0.43, L0.44, L0.45, L0.46, L0.47, L0.48, L0.49, L0.50, L0.51, L0.52, L0.53, L0.54, L0.55, L0.56, L0.57, L0.58, L0.59, L0.60, L0.61, L0.62, L0.63, L0.64, L0.65, L0.66, L0.67, L0.68, L0.69, L0.70, L0.71, L0.72, L0.73, L0.74, L0.75, L0.76, L0.77, L0.78, L0.79, L0.80, L0.81, L0.82, L0.83, L0.84, L0.85, L0.86, L0.87, L0.88, L0.89, L0.90, L0.91, L0.92, L0.93, L0.94, L0.95, L0.96, L0.97, L0.98, L0.99, L0.100, L0.101, L0.102, L0.103, L0.104, L0.105, L0.106, L0.107, L0.108, L0.109, L0.110, L0.111, L0.112, L0.113, L0.114, L0.115, L0.116, L0.117, L0.118, L0.119, L0.120, L0.121, L0.122, L0.123, L0.124, L0.125, L0.126, L0.127, L0.128, L0.129, L0.130, L0.131, L0.132, L0.133, L0.134, L0.135, L0.136, L0.137, L0.138, L0.139, L0.140, L0.141, L0.142, L0.143, L0.144, L0.145, L0.146, L0.147, L0.148, L0.149, L0.150"
    - "  Creating 3 files"
    - "**** Simulation run 1, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[151, 201]). 101 Input Files, 299mb total:"
    - "L0                                                                                                                 "
    - "L0.151[150,151] 151ns 2mb                             |L0.151|                                                     "
    - "L0.152[151,152] 152ns 2mb                              |L0.152|                                                    "
    - "L0.153[152,153] 153ns 2mb                              |L0.153|                                                    "
    - "L0.154[153,154] 154ns 2mb                               |L0.154|                                                   "
    - "L0.155[154,155] 155ns 2mb                                |L0.155|                                                  "
    - "L0.156[155,156] 156ns 2mb                                |L0.156|                                                  "
    - "L0.157[156,157] 157ns 2mb                                 |L0.157|                                                 "
    - "L0.158[157,158] 158ns 2mb                                 |L0.158|                                                 "
    - "L0.159[158,159] 159ns 2mb                                  |L0.159|                                                "
    - "L0.160[159,160] 160ns 2mb                                   |L0.160|                                               "
    - "L0.161[160,161] 161ns 2mb                                   |L0.161|                                               "
    - "L0.162[161,162] 162ns 2mb                                    |L0.162|                                              "
    - "L0.163[162,163] 163ns 2mb                                    |L0.163|                                              "
    - "L0.164[163,164] 164ns 2mb                                     |L0.164|                                             "
    - "L0.165[164,165] 165ns 2mb                                      |L0.165|                                            "
    - "L0.166[165,166] 166ns 2mb                                      |L0.166|                                            "
    - "L0.167[166,167] 167ns 2mb                                       |L0.167|                                           "
    - "L0.168[167,168] 168ns 2mb                                       |L0.168|                                           "
    - "L0.169[168,169] 169ns 2mb                                        |L0.169|                                          "
    - "L0.170[169,170] 170ns 2mb                                         |L0.170|                                         "
    - "L0.171[170,171] 171ns 2mb                                         |L0.171|                                         "
    - "L0.172[171,172] 172ns 2mb                                          |L0.172|                                        "
    - "L0.173[172,173] 173ns 2mb                                          |L0.173|                                        "
    - "L0.174[173,174] 174ns 2mb                                           |L0.174|                                       "
    - "L0.175[174,175] 175ns 2mb                                            |L0.175|                                      "
    - "L0.176[175,176] 176ns 2mb                                            |L0.176|                                      "
    - "L0.177[176,177] 177ns 2mb                                             |L0.177|                                     "
    - "L0.178[177,178] 178ns 2mb                                             |L0.178|                                     "
    - "L0.179[178,179] 179ns 2mb                                              |L0.179|                                    "
    - "L0.180[179,180] 180ns 2mb                                               |L0.180|                                   "
    - "L0.181[180,181] 181ns 2mb                                               |L0.181|                                   "
    - "L0.182[181,182] 182ns 2mb                                                |L0.182|                                  "
    - "L0.183[182,183] 183ns 2mb                                                |L0.183|                                  "
    - "L0.184[183,184] 184ns 2mb                                                 |L0.184|                                 "
    - "L0.185[184,185] 185ns 2mb                                                  |L0.185|                                "
    - "L0.186[185,186] 186ns 2mb                                                  |L0.186|                                "
    - "L0.187[186,187] 187ns 2mb                                                   |L0.187|                               "
    - "L0.188[187,188] 188ns 2mb                                                   |L0.188|                               "
    - "L0.189[188,189] 189ns 2mb                                                    |L0.189|                              "
    - "L0.190[189,190] 190ns 2mb                                                     |L0.190|                             "
    - "L0.191[190,191] 191ns 2mb                                                     |L0.191|                             "
    - "L0.192[191,192] 192ns 2mb                                                      |L0.192|                            "
    - "L0.193[192,193] 193ns 2mb                                                      |L0.193|                            "
    - "L0.194[193,194] 194ns 2mb                                                       |L0.194|                           "
    - "L0.195[194,195] 195ns 2mb                                                        |L0.195|                          "
    - "L0.196[195,196] 196ns 2mb                                                        |L0.196|                          "
    - "L0.197[196,197] 197ns 2mb                                                         |L0.197|                         "
    - "L0.198[197,198] 198ns 2mb                                                         |L0.198|                         "
    - "L0.199[198,199] 199ns 2mb                                                          |L0.199|                        "
    - "L0.200[199,200] 200ns 2mb                                                           |L0.200|                       "
    - "L0.201[200,201] 201ns 2mb                                                           |L0.201|                       "
    - "L0.202[201,202] 202ns 2mb                                                            |L0.202|                      "
    - "L0.203[202,203] 203ns 2mb                                                             |L0.203|                     "
    - "L0.204[203,204] 204ns 2mb                                                             |L0.204|                     "
    - "L0.205[204,205] 205ns 2mb                                                              |L0.205|                    "
    - "L0.206[205,206] 206ns 2mb                                                              |L0.206|                    "
    - "L0.207[206,207] 207ns 2mb                                                               |L0.207|                   "
    - "L0.208[207,208] 208ns 2mb                                                                |L0.208|                  "
    - "L0.209[208,209] 209ns 2mb                                                                |L0.209|                  "
    - "L0.210[209,210] 210ns 2mb                                                                 |L0.210|                 "
    - "L0.211[210,211] 211ns 2mb                                                                 |L0.211|                 "
    - "L0.212[211,212] 212ns 2mb                                                                  |L0.212|                "
    - "L0.213[212,213] 213ns 2mb                                                                   |L0.213|               "
    - "L0.214[213,214] 214ns 2mb                                                                   |L0.214|               "
    - "L0.215[214,215] 215ns 2mb                                                                    |L0.215|              "
    - "L0.216[215,216] 216ns 2mb                                                                    |L0.216|              "
    - "L0.217[216,217] 217ns 2mb                                                                     |L0.217|             "
    - "L0.218[217,218] 218ns 2mb                                                                      |L0.218|            "
    - "L0.219[218,219] 219ns 2mb                                                                      |L0.219|            "
    - "L0.220[219,220] 220ns 2mb                                                                       |L0.220|           "
    - "L0.221[220,221] 221ns 2mb                                                                       |L0.221|           "
    - "L0.222[221,222] 222ns 2mb                                                                        |L0.222|          "
    - "L0.223[222,223] 223ns 2mb                                                                         |L0.223|         "
    - "L0.224[223,224] 224ns 2mb                                                                         |L0.224|         "
    - "L0.225[224,225] 225ns 2mb                                                                          |L0.225|        "
    - "L0.226[225,226] 226ns 2mb                                                                          |L0.226|        "
    - "L0.227[226,227] 227ns 2mb                                                                           |L0.227|       "
    - "L0.228[227,228] 228ns 2mb                                                                            |L0.228|      "
    - "L0.229[228,229] 229ns 2mb                                                                            |L0.229|      "
    - "L0.230[229,230] 230ns 2mb                                                                             |L0.230|     "
    - "L0.231[230,231] 231ns 2mb                                                                             |L0.231|     "
    - "L0.232[231,232] 232ns 2mb                                                                              |L0.232|    "
    - "L0.233[232,233] 233ns 2mb                                                                               |L0.233|   "
    - "L0.234[233,234] 234ns 2mb                                                                               |L0.234|   "
    - "L0.235[234,235] 235ns 2mb                                                                                |L0.235|  "
    - "L0.236[235,236] 236ns 2mb                                                                                |L0.236|  "
    - "L0.237[236,237] 237ns 2mb                                                                                 |L0.237| "
    - "L0.238[237,238] 238ns 2mb                                                                                  |L0.238|"
    - "L0.239[238,239] 239ns 2mb                                                                                  |L0.239|"
    - "L0.240[239,240] 240ns 2mb                                                                                   |L0.240|"
    - "L0.241[240,241] 241ns 2mb                                                                                   |L0.241|"
    - "L0.242[241,242] 242ns 2mb                                                                                    |L0.242|"
    - "L0.243[242,243] 243ns 2mb                                                                                     |L0.243|"
    - "L0.244[243,244] 244ns 2mb                                                                                     |L0.244|"
    - "L0.245[244,245] 245ns 2mb                                                                                      |L0.245|"
    - "L0.246[245,246] 246ns 2mb                                                                                      |L0.246|"
    - "L0.247[246,247] 247ns 2mb                                                                                       |L0.247|"
    - "L0.248[247,248] 248ns 2mb                                                                                        |L0.248|"
    - "L0.249[248,249] 249ns 2mb                                                                                        |L0.249|"
    - "L0.250[249,250] 250ns 2mb                                                                                         |L0.250|"
    - "L1                                                                                                                 "
    - "L1.291[101,150] 150ns 99mb|----------L1.291-----------|                                                             "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 299mb total:"
    - "L1                                                                                                                 "
    - "L1.?[101,151] 250ns 102mb|------------L1.?------------|                                                            "
    - "L1.?[152,201] 250ns 100mb                              |-----------L1.?------------|                               "
    - "L1.?[202,250] 250ns 98mb                                                              |-----------L1.?-----------| "
    - "Committing partition 1:"
    - "  Soft Deleting 101 files: L0.151, L0.152, L0.153, L0.154, L0.155, L0.156, L0.157, L0.158, L0.159, L0.160, L0.161, L0.162, L0.163, L0.164, L0.165, L0.166, L0.167, L0.168, L0.169, L0.170, L0.171, L0.172, L0.173, L0.174, L0.175, L0.176, L0.177, L0.178, L0.179, L0.180, L0.181, L0.182, L0.183, L0.184, L0.185, L0.186, L0.187, L0.188, L0.189, L0.190, L0.191, L0.192, L0.193, L0.194, L0.195, L0.196, L0.197, L0.198, L0.199, L0.200, L0.201, L0.202, L0.203, L0.204, L0.205, L0.206, L0.207, L0.208, L0.209, L0.210, L0.211, L0.212, L0.213, L0.214, L0.215, L0.216, L0.217, L0.218, L0.219, L0.220, L0.221, L0.222, L0.223, L0.224, L0.225, L0.226, L0.227, L0.228, L0.229, L0.230, L0.231, L0.232, L0.233, L0.234, L0.235, L0.236, L0.237, L0.238, L0.239, L0.240, L0.241, L0.242, L0.243, L0.244, L0.245, L0.246, L0.247, L0.248, L0.249, L0.250, L1.291"
    - "  Creating 3 files"
    - "**** Simulation run 2, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[252]). 39 Input Files, 174mb total:"
    - "L0                                                                                                                 "
    - "L0.288[287,288] 288ns 2mb                                                                                        |L0.288|"
    - "L0.287[286,287] 287ns 2mb                                                                                       |L0.287|"
    - "L0.286[285,286] 286ns 2mb                                                                                      |L0.286|"
    - "L0.285[284,285] 285ns 2mb                                                                                     |L0.285|"
    - "L0.284[283,284] 284ns 2mb                                                                                    |L0.284|"
    - "L0.283[282,283] 283ns 2mb                                                                                   |L0.283|"
    - "L0.282[281,282] 282ns 2mb                                                                                  |L0.282|"
    - "L0.281[280,281] 281ns 2mb                                                                                 |L0.281| "
    - "L0.280[279,280] 280ns 2mb                                                                                |L0.280|  "
    - "L0.279[278,279] 279ns 2mb                                                                               |L0.279|   "
    - "L0.278[277,278] 278ns 2mb                                                                              |L0.278|    "
    - "L0.277[276,277] 277ns 2mb                                                                             |L0.277|     "
    - "L0.276[275,276] 276ns 2mb                                                                            |L0.276|      "
    - "L0.275[274,275] 275ns 2mb                                                                           |L0.275|       "
    - "L0.274[273,274] 274ns 2mb                                                                          |L0.274|        "
    - "L0.273[272,273] 273ns 2mb                                                                         |L0.273|         "
    - "L0.272[271,272] 272ns 2mb                                                                        |L0.272|          "
    - "L0.271[270,271] 271ns 2mb                                                                       |L0.271|           "
    - "L0.270[269,270] 270ns 2mb                                                                      |L0.270|            "
    - "L0.269[268,269] 269ns 2mb                                                                     |L0.269|             "
    - "L0.268[267,268] 268ns 2mb                                                                    |L0.268|              "
    - "L0.267[266,267] 267ns 2mb                                                                  |L0.267|                "
    - "L0.266[265,266] 266ns 2mb                                                                 |L0.266|                 "
    - "L0.265[264,265] 265ns 2mb                                                                |L0.265|                  "
    - "L0.264[263,264] 264ns 2mb                                                               |L0.264|                   "
    - "L0.263[262,263] 263ns 2mb                                                              |L0.263|                    "
    - "L0.262[261,262] 262ns 2mb                                                             |L0.262|                     "
    - "L0.261[260,261] 261ns 2mb                                                            |L0.261|                      "
    - "L0.260[259,260] 260ns 2mb                                                           |L0.260|                       "
    - "L0.259[258,259] 259ns 2mb                                                          |L0.259|                        "
    - "L0.258[257,258] 258ns 2mb                                                         |L0.258|                         "
    - "L0.257[256,257] 257ns 2mb                                                        |L0.257|                          "
    - "L0.256[255,256] 256ns 2mb                                                       |L0.256|                           "
    - "L0.255[254,255] 255ns 2mb                                                      |L0.255|                            "
    - "L0.254[253,254] 254ns 2mb                                                     |L0.254|                             "
    - "L0.253[252,253] 253ns 2mb                                                    |L0.253|                              "
    - "L0.252[251,252] 252ns 2mb                                                   |L0.252|                               "
    - "L0.251[250,251] 251ns 2mb                                                  |L0.251|                                "
    - "L1                                                                                                                 "
    - "L1.294[202,250] 250ns 98mb|---------------------L1.294---------------------|                                        "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 174mb total:"
    - "L1                                                                                                                 "
    - "L1.?[202,252] 288ns 102mb|-----------------------L1.?-----------------------|                                      "
    - "L1.?[253,288] 288ns 72mb                                                      |---------------L1.?---------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 39 files: L0.251, L0.252, L0.253, L0.254, L0.255, L0.256, L0.257, L0.258, L0.259, L0.260, L0.261, L0.262, L0.263, L0.264, L0.265, L0.266, L0.267, L0.268, L0.269, L0.270, L0.271, L0.272, L0.273, L0.274, L0.275, L0.276, L0.277, L0.278, L0.279, L0.280, L0.281, L0.282, L0.283, L0.284, L0.285, L0.286, L0.287, L0.288, L1.294"
    - "  Creating 2 files"
    - "**** Simulation run 3, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[101]). 2 Input Files, 201mb total:"
    - "L1                                                                                                                 "
    - "L1.290[51,100] 150ns 99mb|------------------L1.290------------------|                                              "
    - "L1.292[101,151] 250ns 102mb                                             |------------------L1.292-------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 201mb total:"
    - "L2                                                                                                                 "
    - "L2.?[51,101] 250ns 102mb |-------------------L2.?--------------------|                                             "
    - "L2.?[102,151] 250ns 100mb                                             |-------------------L2.?-------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.290, L1.292"
    - "  Upgrading 1 files level to CompactionLevel::L2: L1.289"
    - "  Creating 2 files"
    - "**** Simulation run 4, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[202, 252]). 3 Input Files, 274mb total:"
    - "L1                                                                                                                 "
    - "L1.296[253,288] 288ns 72mb                                                                  |-------L1.296--------| "
    - "L1.293[152,201] 250ns 100mb|------------L1.293------------|                                                          "
    - "L1.295[202,252] 288ns 102mb                                 |------------L1.295-------------|                        "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 274mb total:"
    - "L2                                                                                                                 "
    - "L2.?[152,202] 288ns 102mb|-------------L2.?--------------|                                                         "
    - "L2.?[203,252] 288ns 100mb                                 |-------------L2.?-------------|                         "
    - "L2.?[253,288] 288ns 72mb                                                                   |--------L2.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.293, L1.295, L1.296"
    - "  Creating 3 files"
    - "**** Final Output Files (1.22gb written)"
    - "L2                                                                                                                 "
    - "L2.289[0,50] 150ns 101mb |---L2.289----|                                                                           "
    - "L2.297[51,101] 250ns 102mb               |---L2.297----|                                                            "
    - "L2.298[102,151] 250ns 100mb                               |---L2.298----|                                            "
    - "L2.299[152,202] 288ns 102mb                                               |---L2.299----|                            "
    - "L2.300[203,252] 288ns 100mb                                                               |---L2.300----|            "
    - "L2.301[253,288] 288ns 72mb                                                                               |-L2.301-| "
    "###
    );
}
